# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

import json
import asyncio
from copy import deepcopy
import re

from foglamp.common.common import _FOGLAMP_ROOT, _FOGLAMP_DATA
from foglamp.plugins.north.aws_iot.awsclient import AwsClient
from foglamp.plugins.north.aws_iot.awsthing import AwsThing
from foglamp.plugins.north.aws_iot.awsmqtt import _LOGGER

__author__ = "Himanshu Vimal"
__copyright__ = "Copyright (c) 2023 Dianomic Systems Inc."
__license__ = "Apache 2.0"


_DEFAULT_CONFIG = {
    'plugin': {
         'description': 'An AWS IoT Core plugin that will send readings to IoT Core and create IoT Core Things as required',
         'type': 'string',
         'default': 'aws_iot',
         'readonly': 'true'
    },
    'awsIotEndpoint': {
        'description': 'AWS IoT endpoint for data communication',
        'type': 'string',
        'default': '',
        'order': '1',
        'displayName': 'AWS IoT Endpoint',
    },
    'attachBillingGroup': {
        'description': 'Attach billing group to things created/updated by plugin.',
        'type': 'boolean',
        'default': 'false',
        'order': '2',
        'displayName': 'Attach Billing Group'
    },
    'billingGroupName': {
        'description': 'The name of the billing group the thing will be added to.',
        'type': 'string',
        'default': '',
        'order': '3',
        'validity': "attachBillingGroup == \"true\"",
        'displayName': 'Billing Group'
    },
    'source': {
        'description': 'Source of data to be sent on the AWS IoT. May be either readings or statistics',
        'type': 'enumeration',
        'default': 'readings',
        'options': ['readings', 'statistics'],
        'order': '4',
        'displayName': 'Source'
    },
    'privateKeyName': {
        'description': 'Name of private key provided by AWS IoT for communication.',
        'type': 'string',
        'default': '',
        'order': '5',
        'displayName': 'Private Key Name',
        'group': 'AWS Device Certificates',
    },
    'certName': {
        'description': 'Name of certificate provided by AWS IoT for device identity authentication.',
        'type': 'string',
        'default': '',
        'order': '6',
        'displayName': 'Certificate Name',
        'group': 'AWS Device Certificates',
    },
    'caCertName': {
        'description': 'Name of the root CA generated by AWS IoT',
        'type': 'string',
        'default': '',
        'order': '7',
        'displayName': 'Root CA Name',
        'group': 'AWS Device Certificates',
    },
    'awsAccessId': {
        'description': 'AWS access key ID attached to IAM role.',
        'type': 'string',
        'default': '',
        'order': '8',
        'displayName': 'AWS Access Key ID',
        'group': 'AWS Access Keys',
    },
    'awsSecretKey': {
        'description': 'AWS Secret Access Key attached to IAM Role',
        'type': 'password',
        'default': '',
        'order': '9',
        'displayName': 'AWS Secret Access Key',
        'group': 'AWS Access Keys',
    },
    'mqttQos': {
        'description': 'QoS policy for MQTT publisher',
        'type': 'enumeration',
        'default': 'AT LEAST ONCE',
        'options': ['AT LEAST ONCE', 'AT MOST ONCE'],
        'order': '10',
        'displayName': 'QoS Level',
        'group': 'MQTT Configuration',
    },
    'keepAliveInterval': {
        'description': 'Keep alive interval for MQTT connection in seconds.',
        'type': 'integer',
        'default': '15',
        'minimum' : '5', # must be greater that ping_timeout (default 3 seconds)
        'order': '11',
        'displayName': 'Keep Alive Interval',
        'group': 'MQTT Configuration',
    },
    # TODO: add proxy options
}

def plugin_info():
    """ Function to provide plugin information during discovery.   
        Returns: 
            Information about the plugin including the configuration for the plugin.
    """
    return {
        'name': 'AWS IoT Core',
        'version': '2.5.0',
        'type': 'north',
        'interface': '2.0.0',
        'config': _DEFAULT_CONFIG
    }

def plugin_init(data):
    """ Interface for initialization of a plugin.
        Args:
            data: Plugin configuration.

        Returns:
            config_data: Dictionary of a plugin configuration.
    """
    config_data = deepcopy(data)
    aws_client = AwsClient()
    aws_client.clean()
    aws_client = AwsClient()
    aws_client.set_config(config_data)
    return config_data

def plugin_start(handle):
    """ Placeholder for plugin_start entrypoint.
        Args: 
            handle: handle returned by plugin initialized call.
    """
    pass

def plugin_shutdown(handle):
    """ Shutdowns the plugin. Closes MQTT client connection and cleans up resources.
        Args:
            handle: handle returned by plugin initialized call.
    """
    _LOGGER.info('{} north plugin shut down'.format(_DEFAULT_CONFIG['plugin']['default']))
    # stop mqtt client, close connection and free up resources
    aws_client = AwsClient()
    if aws_client.mqtt_client:
        aws_client.mqtt_client.disconnect()
        
    aws_client.clean()


async def plugin_send(handle, reading_payload, stream_id):
    """ Plugin Interface to communicate with North service.
        Args:
            handle: An object returned by plugin_init.
            reading_payload: Block of readings sent by North service.
            stream_id: Integer that uniquely identifies connection b/w foglamp instance and destination.
        
        Returns:
            Tuple which consists of:
            is_data_sent: A Boolean that indicates if any data has been sent.
            last_object_id: The object id of the last reading which has been sent.
            num_sent: Total number of readings which has been sent to the configured destination.
    """   
    try:
        aws_client = AwsClient()
        is_data_sent = False
        last_object_id = reading_payload[0]['id']
        num_sent = 0
        # parse and send readings        
        for p in reading_payload:
            last_object_id = p['id']

            # #initalize mqtt connection if running first time
            if aws_client.mqtt_client is None:
                aws_client.configure_mqtt(handle)
                aws_client.mqtt_client.create_connection()

            # check mqtt connection
            if aws_client.mqtt_client.is_connected == False: 
                aws_client.mqtt_client.start_publisher()
                break
            # check for valid billing group
            if not aws_client.billing_group_exist:
                break
            await _publish_reading(p)
            # update sent count
            num_sent += 1
            is_data_sent = True
    except asyncio.CancelledError:
        pass
    except Exception as ex:
        _LOGGER.exception(ex, 'Failed to publish readings on AWS IoT. Check connectivity to AWS IoT')
    else:
        _LOGGER.debug("data sent {} last object id {} num sent {}".format(is_data_sent, last_object_id, num_sent))
        return is_data_sent, last_object_id, num_sent

async def _publish_reading(reading):
    """ Publish Reading to AWS IoT
        Args: 
            thing_config: Dict containing required arguments to create/update a thing.
            aws_hint: Datapoint received in reading containing 'Thing' hints.
    """
    aws_client = AwsClient()

    #extract datapoints and AWSHint from readings
    msg_to_publish = dict()
    aws_hint = None
    kwargs = dict()
    kwargs['thingName'] = reading['asset_code']
    kwargs['thingTypeName'] = 'FogLAMPAsset'    # default asset type
    kwargs['topic'] = kwargs['thingName']

    for k, v in reading['reading'].items():
        if k == 'AWSHint':
            try:
                aws_hint = json.loads(v)
            except:
                _LOGGER.warning("Ignoring invalid AWSHint JSON format. Reconfigure AWSHint filter with correct JSON")
            continue
        msg_to_publish[k] = v
    
    msg_to_publish['timestamp'] = reading['user_ts']
    msg_to_publish['asset'] = kwargs['thingName']
    
    if aws_hint:
        extract_hint_info(kwargs, aws_hint)
    topic_name = kwargs.pop('topic')
    #sanitize thing name
    kwargs['thingName'] = sanitize_thing_name(kwargs['thingName'])
    # create/update thing
    update_thing(kwargs)

    # publish message
    await aws_client.mqtt_client.publish(topic=topic_name, message=msg_to_publish)

def update_thing(thing_config):
    """ Create/Update 'Thing' on AWS IoT with information extracted from readings.
        Args: 
            thing_config: Dict containing required arguments to create/update a thing.
    """
    aws_client = AwsClient()
    # check if thing exists
    if not aws_client.thing_exists(name=thing_config['thingName']):
        # create thing, handle error and retry
        created = aws_client.create_thing(**thing_config)
        if not created:
            return
    else:
        # check for thing update
        update_req = False
        new_thing = AwsThing()
        new_thing.set_properties(**thing_config)
        if not aws_client.name_to_awsthing[thing_config['thingName']] == new_thing:
            update_req = True
        if update_req:
            updated = aws_client.update_thing(**thing_config)
            if not updated:
                # log warning, but continue sending reading
                _LOGGER.warning("Check provided AWS Credentials are authorized for update operation and AWSHint format is correct")
            else:
                aws_client.name_to_awsthing[thing_config['thingName']] = new_thing

def extract_hint_info(thing_config, aws_hint):
    """ Updates keyword arguments required to create/update a thing.
        Args: 
            thing_config: Dict containing required arguments to create/update a thing.
            aws_hint: Datapoint received in reading containing 'Thing' hints.
    """
    for key in aws_hint:
        if key == 'name':
            thing_config['thingName'] = aws_hint['name']
            thing_config['topic'] = aws_hint['name']  #default topic name
        elif key == 'type':
            thing_config['thingTypeName'] = aws_hint['type']
        elif key == 'attributes':
            thing_config['attributePayload'] = {'attributes' : aws_hint['attributes']}
        elif key == 'topic':
            thing_config['topic'] = aws_hint['topic']

def sanitize_thing_name(asset_name):
    """ Removes characters not matching valid pattern for thingName.
        Args: 
            asset_name: Asset code string from a reading.            
        Returns:
            sanitized_name: String replaced with '-' for non-matching char.
    """
    sanitized_name = re.sub('[^a-zA-Z0-9:_-]+', '-', asset_name)
    return sanitized_name

