# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

from awscrt import io, mqtt
from awscrt.exceptions import AwsCrtError
from awsiot import mqtt_connection_builder
from threading import Thread
import asyncio
import time as t
import uuid
import os
import json

from foglamp.common.common import _FOGLAMP_ROOT, _FOGLAMP_DATA
from foglamp.common import logger

_LOGGER = logger.setup(__name__)

event_loop_group = None
host_resolver = None
client_bootstrap = None

class AwsMqtt(object):
    """ Mqtt object to publish message to AWS IoT Core
    """
    __slots__ = {'event_loop', 'end_point', 'client_id', 'private_key_path', 'cert_path', 'ca_cert_path', 'is_connected',
                 'mqtt_connection', 'shutdown_connection', 'connect_thread', 'mqtt_publisher_qos',
                  'keep_alive_interval', '__weakref__'}
    
    def __init__(self, config):
        """ Constructor method for AwsMqtt
            Args:
                end_point: Host name of AWS IoT server.
                client_id: An identity placed in MQTT connect packet. It must be unique across all devices/clients.
                private_key_path: File path of AWS IoT private key used to authenticate device.
                cert_path: File path of certificate paired with above private key for authentication.
                ca_cert_path: File path of Certificate Authority to verify certificate.
                is_connected: A bool representing state of mqtt connection.
                mqtt_connection: An object of crt.mqtt.connection representing MQTT connection
                shutdown_connection: A bool to indicate termination of connection                
                connect_thread: A thread to retry MQTT connection.
                mqtt_publisher_qos: Defines QOS level when publishing messages.(0: At most once delivery,
                                                                                1: At least once delivery)
                keep_alive_interval: Keep alive interval for connection in seconds
        """
        self.event_loop = asyncio.get_event_loop()
        self.end_point = None
        self.client_id = str(uuid.getnode())  #unique id based on hardware, uuid.getnode
        self.private_key_path = None
        self.cert_path = None
        self.ca_cert_path = None
        self.is_connected = False
        self.mqtt_connection = None
        self.shutdown_connection = False
        self.connect_thread = None
        self.mqtt_publisher_qos = mqtt.QoS.AT_LEAST_ONCE
        self.keep_alive_interval = 5
        self.set_config(config)

    def set_config(self, config):
        """ Setter for attributes received in plugin configuration.
            Args: 
                config: Dict containing aws endpoint, certifcate, key and other configuration.
        """
        cert_storage_path = self.get_certs_dir()
        self.end_point = config['awsIotEndpoint']['value']
        self.private_key_path = cert_storage_path + config['privateKeyName']['value']
        self.cert_path = cert_storage_path + config['certName']['value']
        self.ca_cert_path = cert_storage_path + 'pem/' + config['caCertName']['value']
        if config['mqttQos']['value'] == 'AT MOST ONCE':
            self.mqtt_publisher_qos = mqtt.QoS.AT_MOST_ONCE
        self.keep_alive_interval = int(config['keepAliveInterval']['value'])

    def conn_resume_callback(self, connection, **kwargs):
        """ Callback invoked whenever the MQTT connection is automatically resumed.
            Args:
                connection: A MQTT connection object.
                retCode: Return code received from the server.
        """
        self.is_connected = True
        self.mqtt_connection = connection
        _LOGGER.info("Connectivity to AWS IoT restored, MQTT connection re-established")

    def conn_interrupt_callback(self, connection, error):
        """ Callback invoked whenever the MQTT connection is lost.
            The MQTT client will automatically attempt to reconnect.
            Args:
                connection: A MQTT connection object.
                error(awscrt.exceptions.AwsCrtError): An exception that caused loss in connection. 
        """
        self.is_connected = False
        self.mqtt_connection = connection
        _LOGGER.error("MQTT connection lost, errcode '{}'. Check connectivity to AWS IoT, "
                      "trying to re-establish connection".format(error))

    def start_publisher(self):
        """ Starts a thread to establish MQTT connection.
        """
        if self.is_connected:
            return
        elif self.connect_thread:
            return
        elif self.mqtt_connection is None:
            _LOGGER.debug("Unable to create MQTT connection, Check AWS IoT endpoint/ certificates")
            return
        
        _LOGGER.info("Connecting to AWS endpoint with Client ID: {}".format(self.client_id))
        self.connect_thread = Thread(target=self.connect, args=(), name='Connect Thread')
        self.connect_thread.start()

    def create_connection(self):
        """ Create MQTT connection object with configuration received from plugin_init.
        """
        try:
            # create connection object for mqtt publisher
            global event_loop_group, host_resolver, client_bootstrap
            event_loop_group = io.EventLoopGroup(2)
            host_resolver = io.DefaultHostResolver(event_loop_group)
            client_bootstrap = io.ClientBootstrap(event_loop_group, host_resolver)
            self.mqtt_connection = mqtt_connection_builder.mtls_from_path(endpoint=self.end_point,
                                                                 cert_filepath=self.cert_path,
                                                                 pri_key_filepath=self.private_key_path,
                                                                 ca_filepath=self.ca_cert_path,
                                                                 client_bootstrap=client_bootstrap,
                                                                 client_id=self.client_id,
                                                                 clean_session=False,
                                                                 keep_alive_secs=self.keep_alive_interval,
                                                                 on_connection_interrupted=self.conn_interrupt_callback,
                                                                 on_connection_resumed=self.conn_resume_callback)
        except FileNotFoundError as ex:
            self.mqtt_connection = None
            _LOGGER.error("Unable to find certificates for device authentication, upload AWS certifcates via FogLAMP Certificate store "
                          "and reconfigure plugin")
        except AwsCrtError as ex:
            self.mqtt_connection = None
            _LOGGER.exception(ex, "Exception occured while configuring MQTT connection")
            _LOGGER.error("Reconfigure plugin with valid AWS IoT endpoint URL and certificates uploaded to FogLAMP Certificate store")


    
    def connect(self):
        """ Establish connection to AWS IoT with MQTT connection object. Retries until plugin shutdown or
            successful connection.
        """
        retry_connection = False
        if self.mqtt_connection is None:
            return
        
        while True:            
            try:
                future =  self.mqtt_connection.connect()
                future.result()
                self.is_connected = True
                _LOGGER.info("MQTT connection established with AWS IoT")
                break
            except AwsCrtError as ex:
                if not retry_connection:
                    _LOGGER.exception(ex, "Exception occured while establishing MQTT connection, "
                                      "check AWS endpoint URL/key/certificates in plugin configuration")
            if self.shutdown_connection == True:
                break
            if not retry_connection:
                retry_connection = True
                _LOGGER.info("Retrying to establish MQTT connection with AWS IoT")
            # re-try after 1 seconds
            t.sleep(1)
            

    def disconnect(self):
        """ Function to disconnect MQTT connection between AWS IoT and device.
        """
        self.shutdown_connection = True

        if self.mqtt_connection:
            self.is_connected = False
            self.shutdown_connection = True
            _LOGGER.info("Plugin shutdown initiated, disconnecting MQTT connection")
            future = self.mqtt_connection.disconnect()
            future.result()
            _LOGGER.info("AWS IoT MQTT client disconnected")
        self.connect_thread = None

    async def publish(self, topic, message):
        """ Publish message to AWS IoT.
            Args:
                topic: Topic Name.
                message: Payload to be sent.
        """
        try:
            payload = json.dumps(message)
        except ValueError as vex:
            _LOGGER.warning("Unable to serialize reading datapoints for topic {}, Skipping non-serializable reading".format(topic))
        else:
            future, _ = (self.mqtt_connection.publish(topic=topic, payload=payload, qos=self.mqtt_publisher_qos))
            future.result()
        
    def get_certs_dir(self):
        """ Helper function to fetch FOGLAMP_DATA directory path.
            Returns:
                Directory path of stored certificates.
        """
        dir_path = _FOGLAMP_DATA + '/etc/certs/' if _FOGLAMP_DATA else _FOGLAMP_ROOT + '/data/etc/certs/'
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        certs_dir = os.path.expanduser(dir_path)
        return certs_dir
    
    